package com.backend.web.ibdtool.controller;

import com.backend.web.ibdtool.entity.Job;
import com.backend.web.ibdtool.exceptions.PatientsNotFoundException;
import com.backend.web.ibdtool.exceptions.QuestionnaireNotFoundException;
import com.backend.web.ibdtool.exceptions.RequestNotAuthorizedException;
import com.backend.web.ibdtool.exceptions.UserNotFoundException;
import com.backend.web.ibdtool.model.BodyStat;
import com.backend.web.ibdtool.model.Patient;
import com.backend.web.ibdtool.services.QuestionnaireService;
import com.backend.web.ibdtool.services.UserService;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.http.ResponseEntity;
import org.springframework.security.access.prepost.PreAuthorize;
import org.springframework.ui.Model;
import org.springframework.web.bind.annotation.*;

import java.io.IOException;
import java.util.ArrayList;


@RestController
public class DoctorController {

    @Autowired
    private UserService userService;

    @Autowired
    private QuestionnaireService questionnaireService;

    @PreAuthorize("hasAnyAuthority('DOTTORE')")
    @GetMapping("private/doctor/get-global-patients/{doctorID}")
    public ResponseEntity<ArrayList<Patient>> getGlobalPatients(@PathVariable String doctorID, Model model) throws UserNotFoundException, PatientsNotFoundException, RequestNotAuthorizedException {

        model.addAttribute(doctorID);

        ArrayList<Patient> patients = userService.getGlobalPatients(doctorID);

        if (patients != null)
            return ResponseEntity.ok(patients);

        return ResponseEntity.badRequest().build();
    }

    @PreAuthorize("hasAnyAuthority('DOTTORE')")
    @GetMapping("private/doctor/get-patients/{doctorID}")
    public ResponseEntity<ArrayList<Patient>> getPatients(@PathVariable String doctorID, Model model) throws UserNotFoundException, PatientsNotFoundException, RequestNotAuthorizedException {

        model.addAttribute(doctorID);

        ArrayList<Patient> patients = userService.getAllPatients(doctorID);

        if (patients != null)
            return ResponseEntity.ok(patients);

        return ResponseEntity.badRequest().build();
    }

    @PreAuthorize("hasAnyAuthority('DOTTORE')")
    @PostMapping("private/doctor/remove-patient/{doctorID}/{email}")
    public ResponseEntity<String> removePatient(@PathVariable String doctorID,
                                                @PathVariable String email,
                                                Model model) throws UserNotFoundException, PatientsNotFoundException, RequestNotAuthorizedException {

        model.addAttribute(doctorID);
        model.addAttribute(email);  // patient that needs to be removed

        userService.removePatient(doctorID, email);

        return ResponseEntity.ok("Patient removed");
    }

    @PreAuthorize("hasAnyAuthority('DOTTORE')")
    @PostMapping("private/doctor/set-questionnaires/{doctorID}/{patientSSN}")
    public ResponseEntity<String> setQuestionnaires(@PathVariable String doctorID,
                                                    @PathVariable String patientSSN,
                                                    @RequestBody String[] questionnaires,
                                                    Model model) throws IOException, QuestionnaireNotFoundException, RequestNotAuthorizedException, InterruptedException {

        model.addAttribute(patientSSN);

        questionnaireService.setQuestionnaires(doctorID, patientSSN, questionnaires);

        return ResponseEntity.ok("Questionnaires added");
    }

    @PreAuthorize("hasAnyAuthority('DOTTORE')")
    @PostMapping("private/doctor/remove-questionnaire/{doctorID}/{patientSSN}")
    public ResponseEntity<String> removeQuestionnaire(@PathVariable String doctorID,
                                                    @PathVariable String patientSSN,
                                                    @RequestBody String questionnaire,
                                                    Model model) throws IOException, QuestionnaireNotFoundException, RequestNotAuthorizedException, InterruptedException {

        model.addAttribute(patientSSN);

        questionnaireService.removeQuestionnaire(doctorID, patientSSN, questionnaire);

        return ResponseEntity.ok("Questionnaires added");
    }

    /**
     * Get all jobs for patient
     * @param doctorID
     * @param patientSSN
     * @param model
     * @return
     * @throws UserNotFoundException
     * @throws PatientsNotFoundException
     * @throws RequestNotAuthorizedException
     */
    @PreAuthorize("hasAnyAuthority('DOTTORE')")
    @GetMapping("private/doctor/get-jobs/{doctorID}/{patientSSN}")
    public ResponseEntity<ArrayList<Job>> getJobs(@PathVariable String doctorID,
                                                      @PathVariable String patientSSN,
                                                      Model model) throws UserNotFoundException {

        model.addAttribute(doctorID);
        model.addAttribute(patientSSN);

        ArrayList<Job> jobs = userService.getAllJobsForPatient(doctorID, patientSSN);

        if (jobs != null)
            return ResponseEntity.ok(jobs);

        return ResponseEntity.badRequest().build();
    }

    @PostMapping("private/doctor/edit-patient-clinical-profile/{doctorID}/{patientEmail}")
    public ResponseEntity<BodyStat> editPatientClinicalProfile(@PathVariable String doctorID, @PathVariable String patientEmail, @RequestBody BodyStat bodyStatDTO, Model model) throws IOException, RequestNotAuthorizedException {

        model.addAttribute(doctorID);
        model.addAttribute(patientEmail);

        BodyStat newBodyStat = userService.editPatientClinicalProfile(doctorID, patientEmail, bodyStatDTO);

        if (newBodyStat == null) {
            return ResponseEntity.badRequest().body(null);
        }

        return ResponseEntity.ok(newBodyStat);
    }

    @PostMapping("private/doctor/delete-note/{doctorID}/{patientEmail}")
    public ResponseEntity<String> removePatientNote(@PathVariable String doctorID, @PathVariable String patientEmail, @RequestBody String note, Model model) throws IOException, RequestNotAuthorizedException {

        model.addAttribute(doctorID);
        model.addAttribute(patientEmail);

        userService.removePatientNote(doctorID, patientEmail, note);

        return ResponseEntity.ok("Note removed");
    }
}
