package com.backend.web.ibdtool.firebase;

import com.backend.web.ibdtool.entity.UserEntity;
import com.backend.web.ibdtool.exceptions.UserNotFoundException;
import com.backend.web.ibdtool.model.NotificationModel;
import com.backend.web.ibdtool.repository.UserRepository;
import com.google.auth.oauth2.GoogleCredentials;
import com.google.firebase.FirebaseApp;
import com.google.firebase.FirebaseOptions;
import com.google.firebase.iid.FirebaseInstanceIdException;
import com.google.firebase.messaging.FirebaseMessaging;
import com.google.firebase.messaging.FirebaseMessagingException;
import com.google.firebase.messaging.Message;
import com.google.firebase.messaging.Notification;
import lombok.extern.slf4j.Slf4j;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.core.io.ClassPathResource;
import org.springframework.stereotype.Service;

import javax.annotation.PostConstruct;
import java.io.IOException;
import java.net.Authenticator;
import java.net.HttpURLConnection;
import java.net.URI;
import java.net.URL;
import java.net.http.HttpClient;
import java.net.http.HttpRequest;
import java.net.http.HttpResponse;
import java.time.LocalDateTime;
import java.time.ZoneId;
import java.time.ZonedDateTime;
import java.time.format.DateTimeFormatter;
import java.util.HashSet;
import java.util.Optional;

@Slf4j
@Service
public class NotificationService {

    @Value("${app.firebase-config}")
    private String firebaseConfig;

    @Value("${authorization}")
    private String authorization;

    private FirebaseApp firebaseApp;

    @Autowired
    private UserRepository userRepository;

    @PostConstruct
    private void initialize() {
        try {
            FirebaseOptions options = new FirebaseOptions.Builder()
                    .setCredentials(GoogleCredentials.fromStream(new ClassPathResource(firebaseConfig).getInputStream())).build();

            if (FirebaseApp.getApps().isEmpty()) {
                this.firebaseApp = FirebaseApp.initializeApp(options);
            } else {
                this.firebaseApp = FirebaseApp.getInstance();
            }
        } catch (IOException e) {
            log.error("Create FirebaseApp Error", e);
        }
    }

    public void subscribeToService(SubscriptionRequestDto subscriptionRequestDto) throws UserNotFoundException {

        Optional<UserEntity> userEntityOptional = userRepository.findByEmail(subscriptionRequestDto.getEmail());

        if (userEntityOptional.isEmpty()) {
            throw new UserNotFoundException();
        }

        UserEntity userEntity = userEntityOptional.get();

        HashSet<String> tokens = userEntity.getTokens();
        tokens.add(subscriptionRequestDto.getToken());
        userEntity.setTokens(tokens);

        userRepository.save(userEntity);

    }

    public void unsubscribe(String email, String token) throws UserNotFoundException {
        Optional<UserEntity> optionalUserEntity = userRepository.findByEmail(email);

        if (optionalUserEntity.isEmpty())
            throw new UserNotFoundException();

        UserEntity userEntity = optionalUserEntity.get();

        HashSet<String> tokens = userEntity.getTokens();
        tokens.remove(token);
        userEntity.setTokens(tokens);

        userRepository.save(userEntity);
    }

    /**
     * send notification to user
     * @param userEntity
     * @param subject
     * @param body
     * @throws IOException
     */
    public void sendNotificationToUser(UserEntity userEntity, String subject, String body) throws IOException {
        HashSet<String> tokens = userEntity.getTokens();

        if (tokens.size() > 0) {

            NotificationRequestDto notificationRequestDto = new NotificationRequestDto();
            notificationRequestDto.setBody(body);
            LocalDateTime nd = ZonedDateTime.now(ZoneId.of("Europe/Rome")).toLocalDateTime();
            String title = nd.format(DateTimeFormatter.ofPattern("dd/MM/yyyy")) + " " +
                    (nd.getHour() < 10 ? ("0" + nd.getHour()) : (nd.getHour())) + ":" +
                    (nd.getMinute() < 10 ? ("0" + nd.getMinute()) : (nd.getMinute())) + ":" +
                    (nd.getSecond() < 10 ? ("0" + nd.getSecond()) : (nd.getSecond())) +
                    " " + subject;
            notificationRequestDto.setTitle(title);
            notificationRequestDto.setTargetEmail(userEntity.getEmail());

            for (String t : tokens) {
                notificationRequestDto.setTarget(t);
                this.sendPushNotification(notificationRequestDto);
            }
        }
    }

    public void sendPushNotification(NotificationRequestDto notificationRequestDto) throws IOException {
        if (notificationRequestDto.getTarget() != null) {
            if (!notificationRequestDto.getTarget().equals("") && !notificationRequestDto.getTarget().equals("null")) {
                Message message = Message.builder()
                        .setToken(notificationRequestDto.getTarget())
                        .setNotification(new Notification(notificationRequestDto.getTitle(), notificationRequestDto.getBody()))
                        .putData("content", notificationRequestDto.getTitle())
                        .putData("body", notificationRequestDto.getBody())
                        .build();

                // Crud op done anyway
                Optional<UserEntity> optionalUserEntity =
                        userRepository.findByEmail(notificationRequestDto.getTargetEmail());
                if (optionalUserEntity.isEmpty())
                    throw new UserNotFoundException();

                UserEntity userEntity = optionalUserEntity.get();
                HashSet<NotificationModel> notifications = userEntity.getNotifications();
                notifications.add(NotificationModel.builder()
                        .title(notificationRequestDto.getTitle())
                        .body(notificationRequestDto.getBody())
                        .read(false)
                        .build());
                userEntity.setNotifications(notifications);
                userRepository.save(userEntity);

                String response = null;
                // Exception block for sending push

                URL obj = new URL("https://iid.googleapis.com/iid/info/" + notificationRequestDto.getTarget());
                HttpURLConnection httpURLConnection = (HttpURLConnection) obj.openConnection();
                httpURLConnection.setRequestMethod("GET");
                httpURLConnection.setRequestProperty("Authorization", this.authorization);
                int responseCode = httpURLConnection.getResponseCode();

                System.out.println("-------------------------------------------------------------------");
                System.out.println("responseCode-->" + responseCode);
                System.out.println("-------------------------------------------------------------------");

                if (responseCode == 200) {
                    try {
                        response = FirebaseMessaging.getInstance().send(message);
                        System.out.println("RESPONSE ---> " + response);
                    } catch (FirebaseMessagingException e) {
                        log.error("Fail to send firebase notification", e);
                    } finally {
                        log.info("Message not sent");
                    }
                } else {
                    HashSet<String> targets = userEntity.getTokens();
                    targets.remove(notificationRequestDto.getTarget());
                    userEntity.setTokens(targets);
                    userRepository.save(userEntity);
                }

            } else {
                // Crud op -> delete useless tokens
                Optional<UserEntity> optionalUserEntity =
                userRepository.findByEmail(notificationRequestDto.getTargetEmail());
                if (optionalUserEntity.isEmpty())
                    throw new UserNotFoundException();

                UserEntity userEntity = optionalUserEntity.get();

                HashSet<String> targets = userEntity.getTokens();
                targets.remove(notificationRequestDto.getTarget());
                userEntity.setTokens(targets);
                userRepository.save(userEntity);
            }
        }
    }

    /**
     * Remove notification method
     * @param email
     * @param notificationModel
     * @throws UserNotFoundException
     */
    public void removeNotification(String email, NotificationModel notificationModel) throws UserNotFoundException {
        Optional<UserEntity> userEntityOptional = userRepository.findByEmail(email);
        if (userEntityOptional.isEmpty())
            throw new UserNotFoundException();

        UserEntity userEntity = userEntityOptional.get();
        HashSet<NotificationModel> notificationModelHashSet = userEntity.getNotifications();
        NotificationModel toRemove = null;

        for (NotificationModel n : notificationModelHashSet) {
            if (n.getTitle().equals(notificationModel.getTitle()) &&
                n.getBody().equals(notificationModel.getBody())) {
                toRemove = n;
            }
        }

        if (toRemove != null) {
            notificationModelHashSet.remove(toRemove);
        }

        userEntity.setNotifications(notificationModelHashSet);
        userRepository.save(userEntity);
    }

    /**
     * Clear all notifications for user
     * @param email
     * @throws UserNotFoundException
     */
    public void clearAllNotifications(String email) throws UserNotFoundException {
        Optional<UserEntity> userEntityOptional = userRepository.findByEmail(email);
        if (userEntityOptional.isEmpty())
            throw new UserNotFoundException();

        UserEntity userEntity = userEntityOptional.get();
        //userRepository.delete(userEntity);
        HashSet<NotificationModel> notificationModelHashSet = userEntity.getNotifications();
        notificationModelHashSet.clear();
        userEntity.setNotifications(notificationModelHashSet);
        userRepository.save(userEntity);
    }

    public void readNotification(String email, NotificationModel notificationModel) throws UserNotFoundException {
        Optional<UserEntity> userEntityOptional = userRepository.findByEmail(email);
        if (userEntityOptional.isEmpty())
            throw new UserNotFoundException();

        UserEntity userEntity = userEntityOptional.get();
        HashSet<NotificationModel> notificationModelHashSet = userEntity.getNotifications();
        notificationModelHashSet.remove(notificationModel);
        notificationModel.setRead(true);
        notificationModelHashSet.add(notificationModel);
        userEntity.setNotifications(notificationModelHashSet);
        userRepository.save(userEntity);
    }
}

