package com.backend.web.ibdtool.scheduling;

import com.backend.web.ibdtool.entity.Job;
import com.backend.web.ibdtool.entity.UserEntity;
import com.backend.web.ibdtool.model.FutureModel;
import com.backend.web.ibdtool.repository.JobRepository;
import com.backend.web.ibdtool.repository.UserRepository;
import com.backend.web.ibdtool.services.RunnableService;
import lombok.extern.slf4j.Slf4j;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.context.annotation.Bean;
import org.springframework.context.annotation.ComponentScan;
import org.springframework.context.annotation.Configuration;
import org.springframework.scheduling.annotation.EnableAsync;
import org.springframework.scheduling.annotation.EnableScheduling;
import org.springframework.scheduling.annotation.Scheduled;
import org.springframework.scheduling.annotation.SchedulingConfigurer;
import org.springframework.scheduling.concurrent.ThreadPoolTaskScheduler;
import org.springframework.scheduling.config.ScheduledTaskRegistrar;
import org.springframework.stereotype.Service;

import java.time.LocalDate;
import java.time.LocalDateTime;
import java.time.ZoneOffset;
import java.util.*;
import java.util.concurrent.ScheduledFuture;
import java.util.concurrent.atomic.AtomicReference;

@Configuration
@EnableScheduling
@EnableAsync
@ComponentScan(basePackages = "com.backend.web.ibdtool")
@Service
public class SchedulerService implements SchedulingConfigurer {

    private static Logger LOGGER = LoggerFactory.getLogger(SchedulerService.class);

    @Autowired
    private RunnableService runnableService;

    @Autowired
    private JobRepository jobRepository;

    @Autowired
    private UserRepository userRepository;

    ScheduledTaskRegistrar scheduledTaskRegistrar;

    /*private HashMap<ScheduledFuture, Boolean> activeFutures = new HashMap<>();
    private HashMap<String, FutureModel> jobList = new HashMap<>();*/

    private HashMap<String, FutureModel> activeFutures = new HashMap<>();


    @Bean
    public ThreadPoolTaskScheduler threadPoolTaskScheduler() {

        ThreadPoolTaskScheduler threadPoolTaskScheduler = new ThreadPoolTaskScheduler();
        threadPoolTaskScheduler.setPoolSize(20);
        threadPoolTaskScheduler.setThreadNamePrefix("ThreadPoolTaskScheduler");
        threadPoolTaskScheduler.initialize();

        return threadPoolTaskScheduler;
    }

    @Override
    public void configureTasks(ScheduledTaskRegistrar taskRegistrar) {

        if (scheduledTaskRegistrar == null) {
            scheduledTaskRegistrar = taskRegistrar;
        }

        if (taskRegistrar.getScheduler() == null) {
            taskRegistrar.setScheduler(threadPoolTaskScheduler());
        }

        this.repositoryLookup();

        activeFutures.forEach((id, futureModel) -> {

            ScheduledFuture scheduledFuture = futureModel.getScheduledFuture();

            if (scheduledFuture == null || (scheduledFuture.isCancelled() && futureModel.getIsActive())) {

                // 2 - Split ID to pass runnable parameters in Runnable method
                // 3 - Start job

                AtomicReference<Date> lastExecutionTimeToSave = new AtomicReference<>();

                scheduledFuture = taskRegistrar.getScheduler().schedule(runnableService.newRunnable(id), triggerContext -> {
                    Calendar nextExecutionTime = new GregorianCalendar();
                    Date lastActualExecutionTime = triggerContext.lastActualExecutionTime();
                    if (lastActualExecutionTime == null) {
                        lastActualExecutionTime = this.dbDateLookup(id);
                        if (lastActualExecutionTime == null) {
                            Date date = new Date();
                            date.setHours(0);
                            date.setMinutes(0);
                            date.setSeconds(0);
                            lastActualExecutionTime = date;
                        }
                    }

                    //System.out.println("LAST EXECUTION: " + lastActualExecutionTime);

                    nextExecutionTime.setTime(lastActualExecutionTime);
                    nextExecutionTime.add(Calendar.SECOND, (int) futureModel.getTimer());
                    lastExecutionTimeToSave.set(lastActualExecutionTime);
                    return nextExecutionTime.getTime();
                });

                futureModel.setScheduledFuture(scheduledFuture);
                jobRepository.save(Job.builder()
                        .id(id)
                        .timer(futureModel.getTimer())
                        .lastExecution(lastExecutionTimeToSave.get()).build());
            }

        });

        /*System.out.println("--------Active jobs after configureTasks:");
        activeFutures.forEach((i, futureModel) -> {
            System.out.println(i + " " + futureModel.getScheduledFuture() + " " + futureModel.getIsActive());
        });
        System.out.println("------------------------------------------");*/


    }

    /**
     * get Last execution for job from db
     * @param id
     * @return
     */
    public Date dbDateLookup(String id) {
        Optional<Job> optionalJob = jobRepository.findById(id);
        if (optionalJob.isPresent()) {
            Job job = optionalJob.get();
            return job.getLastExecution();
        } else {
            return null;
        }
    }


    /**
     * Get active jobs from repository in case server shuts down
     */
    public void repositoryLookup() {

        List<Job> jobsFromRepository = new LinkedList<>(jobRepository.findAll());

        jobsFromRepository.forEach(job -> {

            if (!activeFutures.containsKey(job.getId())) {

                ScheduledFuture scheduledFuture = null;

                activeFutures.put(job.getId(),
                        FutureModel.builder()
                                .scheduledFuture(scheduledFuture).isActive(true)
                                .timer(job.getTimer())
                                .lastExecution(job.getLastExecution())
                                .build());
            }

        });

        /*System.out.println("--------Active jobs after Repository lookup:");
        activeFutures.forEach((i, futureModel) -> {
            System.out.println(i + " " + futureModel.getScheduledFuture() + " " + futureModel.getIsActive());
        });
        System.out.println("------------------------------------------");*/
    }

    /**
     * Add a new task
     * @param doctorID
     * @param patientSSN
     * @param type
     * @param timer
     */
    public void startTask(String doctorID, String patientSSN, String type, long timer) {
        Optional<UserEntity> optionalPatient = userRepository.findBySSN(patientSSN);
        assert(optionalPatient.isPresent());
        String realDoctorID = optionalPatient.get().getDoctor().getDoctorID();
        String id = realDoctorID + "/" + patientSSN + "/" + type;
        System.out.println(id);
        ScheduledFuture scheduledFuture = null;
        activateFuture(scheduledFuture, id, timer);
    }

    /**
     * Kill an existent job
     * @param doctorID
     * @param patientSSN
     * @param type
     */
    public void killJob(String doctorID, String patientSSN, String type) {
        Optional<UserEntity> optionalPatient = userRepository.findBySSN(patientSSN);
        assert(optionalPatient.isPresent());
        String realDoctorID = optionalPatient.get().getDoctor().getDoctorID();
        String id = realDoctorID + "/" + patientSSN + "/" + type;
        ScheduledFuture toKill = activeFutures.get(id).getScheduledFuture();
        cancelFuture(true, toKill, id);

        Optional<Job> optionalJob = jobRepository.findById(id);
        if (optionalJob.isEmpty()) {
            throw new RuntimeException();
        }
        Job job = optionalJob.get();

        jobRepository.delete(job);

        //System.out.println("--------Active jobs after killing a job:");
        /*activeFutures.forEach((i, futureModel) -> {
            System.out.println(i + " " + futureModel.getScheduledFuture() + " " + futureModel.getIsActive());
        });*/
        //System.out.println("------------------------------------------");
    }

    /**
     * Kill  job by existing id
     * @param id
     */
    public void killJobById(String id) {
        ScheduledFuture toKill = activeFutures.get(id).getScheduledFuture();
        cancelFuture(true, toKill, id);

        Optional<Job> optionalJob = jobRepository.findById(id);
        if (optionalJob.isEmpty()) {
            throw new RuntimeException();
        }
        Job job = optionalJob.get();

        jobRepository.delete(job);

        /*System.out.println("--------Active jobs after killing a job:");
        activeFutures.forEach((i, futureModel) -> {
            System.out.println(i + " " + futureModel.getScheduledFuture() + " " + futureModel.getIsActive());
        });
        System.out.println("------------------------------------------");*/
    }

    /**
     * Activate future
     * @param future
     * @param id
     */
    public void activateFuture(ScheduledFuture future, String id, long timer) {

        //System.out.println("Activating a future");

        if (activeFutures.containsKey(id)) {
            //System.out.println("Job found --> canceling before activating new one");
            ScheduledFuture active = activeFutures.get(id).getScheduledFuture();
            cancelFuture(true, active, id);
        }


        //jobList.put(id, future);    // new future -> equals null -> not activated
        //activeFutures.put(future, true); -> old version

        activeFutures.put(id, FutureModel.builder()
                .isActive(true)
                .scheduledFuture(future)
                .timer(timer)
                .lastExecution(null)
                .build());

        configureTasks(scheduledTaskRegistrar);
    }

    /**
     * @param mayInterruptIfRunning {@code true} if the thread executing this task
     * should be interrupted; otherwise, in-progress tasks are allowed to complete
     */
    public void cancelFuture(boolean mayInterruptIfRunning, ScheduledFuture future, String id) {
        //System.out.println("Cancelling a future");

        future.cancel(mayInterruptIfRunning); // set to false if you want the running task to be completed first.

        // DEBUG -> Remove job
        activeFutures.remove(id);
        // AFTER TRY
        /*activeFutures.remove(future);
        jobList.remove(id);*/

        /*System.out.println("--------Active jobs after activateFuture:");
        activeFutures.forEach((i, futureModel) -> {
            System.out.println(i + " " + futureModel.getScheduledFuture() + " " + futureModel.getIsActive());
        });
        System.out.println("------------------------------------------");*/
    }

    /**
     * kill all active jobs
     */
    public void cancelAll() {
        //System.out.println("Canceling all futures");
        activeFutures.forEach((id, futureModel) -> {
            cancelFuture(true, futureModel.getScheduledFuture(), id);
        });
       /* System.out.println("Active futures after cancel all:");
        activeFutures.forEach((id, futureModel) -> {
            System.out.println(id + " " + futureModel.getScheduledFuture() + " " + futureModel.getIsActive());
        });*/
    }

}
