package com.backend.web.ibdtool.services;

import com.backend.web.ibdtool.dto.*;
import com.backend.web.ibdtool.entity.Job;
import com.backend.web.ibdtool.entity.Pending;
import com.backend.web.ibdtool.entity.UserEntity;
import com.backend.web.ibdtool.exceptions.*;
import com.backend.web.ibdtool.firebase.NotificationRequestDto;
import com.backend.web.ibdtool.firebase.NotificationService;
import com.backend.web.ibdtool.model.*;
import com.backend.web.ibdtool.repository.JobRepository;
import com.backend.web.ibdtool.repository.PendingRepository;
import com.backend.web.ibdtool.repository.UserRepository;
import com.backend.web.ibdtool.scheduling.SchedulerService;
import com.backend.web.ibdtool.security.HashPwdConfig;
import com.backend.web.ibdtool.security.JwtProvider;

import org.apache.catalina.User;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.security.core.Authentication;
import org.springframework.security.core.context.SecurityContextHolder;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;

import java.io.IOException;
import java.security.NoSuchAlgorithmException;
import java.time.*;
import java.time.format.DateTimeFormatter;
import java.util.*;

@Service
public class UserService {

    private Authentication authentication;

    @Autowired
    private UserRepository userRepository;

    @Autowired
    private PendingRepository pendingRepository;

    @Autowired
    private JobRepository jobRepository;

    @Autowired
    private SchedulerService schedulerService;

    @Autowired
    private QuestionnaireService questionnaireService;

    @Autowired
    private JwtProvider jwtProvider;

    @Autowired
    private MailService mailService;

    @Autowired
    private NotificationService notificationService;

    @Value("${ibd-tool-link}") private String link;
    @Value("${ibd-tool-be}") private String linkBe;
    @Value("${manuale-medico}") private String pathDoctor;
    @Value("${manuale-paziente}") private String pathPatient;
    @Value("${ibd-app-playstore}") private String linkApp;


    /**
     * Registration function for doctor
     * @param registrationDTO
     * @throws UserAlreadyRegisteredException
     */
    @Transactional
    public void register(RegistrationDTO registrationDTO)
            throws UserAlreadyRegisteredException, NoSuchAlgorithmException {

        if (userRepository.findByEmail(registrationDTO.getEmail()).isPresent() ||
                userRepository.findBySSN(registrationDTO.getSSN()).isPresent())
            throw new UserAlreadyRegisteredException();

        String uuid = UUID.randomUUID().toString();
        long expireRegistrationTime = LocalDateTime.now().plusDays(1).atZone(ZoneOffset.UTC).toEpochSecond();

        // ID utente: 5 caratteri -> ottenuto da SHA-256(UUID + email + password + SSN)

        StringBuffer UID = new StringBuffer(uuid+registrationDTO.getEmail()+registrationDTO.getPassword()+registrationDTO.getSSN());
        String doctorID = new HashPwdConfig(UID.toString()).getHash().substring(0, 5);

        // STRATEGY:
        // SE TROVO UN DOCTOR ID UGUALE NEL REPO, RICALCOLO UN NUOVO HASH SULLO USER ID CHE STO CERCANDO
        // ED ESTRAGGO I PRIMI 5 CARATTERI

        while ((userRepository.findByDoctorID(doctorID)).isPresent()) {
            doctorID = new HashPwdConfig(doctorID).getHash().substring(0, 5);
        }

        String[] birth = registrationDTO.getBirthDate().split("/");
        int birthDay = Integer.parseInt(birth[0]);
        int birthMonth = Integer.parseInt(birth[1]);
        int birthYear = Integer.parseInt(birth[2]);

        LocalDate birthDate = LocalDate.of(birthYear, birthMonth, birthDay);
        String birthPlace = registrationDTO.getBirthPlace();

        UserEntity user = UserEntity.builder()
                .email(registrationDTO.getEmail())
                .name(registrationDTO.getName())
                .surname(registrationDTO.getSurname())
                .SSN(registrationDTO.getSSN())
                .birthDate(birthDate)
                .birthPlace(birthPlace)
                .phoneNumber(registrationDTO.getPhoneNumber())
                .password(registrationDTO.getPassword())
                .role(registrationDTO.getRole())
                .doctorID(doctorID)
                .uuid(uuid)
                .registrationTime(expireRegistrationTime)
                .enabled(false)
                .patients(null)
                .doctor(null)
                .bodyStat(null)
                .tokens(null)
                .category(null)
                .notifications(null)
                .build();

        userRepository.save(user);

        String link = linkBe + "/public/confirm/" + uuid;
        String message = "Benvenuto in IBD Tool!\nPer confermare la tua registrazione clicca sul seguente link: " + link + "\n" +
                "Il link ha un giorno di validità. Scaduto tale periodo, sarà necessario procedere con una nuova registrazione.\n" +
                "Cordiali saluti,\n" +
                "IBD Tool - Ospedale Mauriziano";

        mailService.sendMail(registrationDTO.getEmail(), "Conferma registrazione in IBD Tool", message);

    }


    /**
     * Confirm registration
     * @param uuid
     * @throws UserNotFoundException
     */
    public void confirmRegistration(String uuid) throws IOException, NoSuchAlgorithmException {
        long time = LocalDateTime.now().atZone(ZoneOffset.UTC).toEpochSecond();

        Optional<UserEntity> optionalUserEntity = userRepository
                .findByEnabledAndUuidAndRegistrationTimeGreaterThan(false, uuid, time);

        if (optionalUserEntity.isEmpty())
            throw new UserNotFoundException();

        UserEntity userEntity = optionalUserEntity.get();

        String password = userEntity.getPassword();
        String hashedPassword = new HashPwdConfig(userEntity.getPassword().concat(userEntity.getEmail())).getHash();

        userEntity.setRegistrationTime(time);
        userEntity.setUuid(null);
        userEntity.setEnabled(true);
        userEntity.setPassword(hashedPassword);
        userEntity.setTokens(new HashSet<>());
        userEntity.setNotifications(new HashSet<>());

        userRepository.save(userEntity);

        // Send email for Confirmed registration

        String message = "Benvenuto in IBD Tool!\nAccedi al portale per scoprire le funzionalità: " +  link + "\n" +
                "Il tuo ID utente è: " + userEntity.getDoctorID() + "\n" +
                "Le tue credenziali per accedere sono:" + "\n" +
                "email: " + userEntity.getEmail() + "\n" +
                "password: " + password;

        mailService.sendMail(userEntity.getEmail(), "Registrazione confermata", message);

        // Send manual

        String subject = "IBD Tool - Manuale di utilizzo";
        String msg = "Benvenuto in IBD Tool!\nConsulta il nostro manuale al link: " + pathDoctor + "\n" +
                "Cordiali saluti,\n" +
                "IBD Tool - Ospedale Mauriziano";

        mailService.sendMail(userEntity.getEmail(), subject, msg);

        // Send Notification to new Doctor registered

        String t = "IBD Tool - Aggiorna il tuo profilo";
        String body = "Benvenuto in IBD Tool.\n" +
                "A breve riceverai un manuale di istruzioni per l'utilizzo.\n" +
                "Nel frattempo aggiorna il tuo profilo dalle impostazioni personali.";

        NotificationRequestDto notificationRequestDto = new NotificationRequestDto();
        notificationRequestDto.setTarget("-");
        notificationRequestDto.setBody(body);
        LocalDateTime nd = ZonedDateTime.now(ZoneId.of("Europe/Rome")).toLocalDateTime();
        String title = nd.format(DateTimeFormatter.ofPattern("dd/MM/yyyy")) + " " +
                (nd.getHour() < 10 ? ("0" + nd.getHour()) : (nd.getHour())) + ":" +
                (nd.getMinute() < 10 ? ("0" + nd.getMinute()) : (nd.getMinute())) + ":" +
                (nd.getSecond() < 10 ? ("0" + nd.getSecond()) : (nd.getSecond())) +
                " " + t;
        notificationRequestDto.setTitle(title);
        notificationRequestDto.setTargetEmail(userEntity.getEmail());

        notificationService.sendPushNotification(notificationRequestDto);

    }

    /**
     * Request change password
     * @param changePasswordRequestDTO
     * @throws UserNotFoundException
     * @throws UserNotEnabledException
     */
    public void changePasswordRequest(ChangePasswordRequestDTO changePasswordRequestDTO) throws UserNotFoundException, UserNotEnabledException {

        //System.out.println(changePasswordRequestDTO.getEmail() + changePasswordRequestDTO.getSSN());
        Optional<UserEntity> optionalUserEntity =
                userRepository.findByEmailAndSSN(changePasswordRequestDTO.getEmail(), changePasswordRequestDTO.getSSN());

        if (optionalUserEntity.isEmpty())
            throw new UserNotFoundException();

        UserEntity userEntity = optionalUserEntity.get();

        String uuid = UUID.randomUUID().toString();
        userEntity.setUuid(uuid);
        userEntity.setPasswordChangeRequest(LocalDateTime.now().plusDays(1).atZone(ZoneOffset.UTC).toEpochSecond());

        //this.removeUser(userEntity.getEmail());
        userRepository.save(userEntity);

        String email = userEntity.getEmail();
        String subject = "IBD Tool - Richiesta cambio password";
        String msg = "Gentile " + userEntity.getName() + " " + userEntity.getSurname() + "\n" +
                "è stato richiesto un cambio password per il tuo account.\n" +
                "Clicca qui per reimpostare la nuova password: " + link + "/change-password-confirm/" + uuid + "\n" +
                "Il link ha un giorno di validità, scaduto tale periodo sarà necessario inoltrare una nuova richiesta dal portale.\nSe non sei stato tu a inoltrare la richiesta, ti preghiamo di ignorare il messaggio.\n" +
                "Cordiali saluti,\n" +
                "IBD Tool - Ospedale Mauriziano";


        mailService.sendMail(email, subject, msg);

    }

    /**
     *
     * @param passwordChangeDTO
     * @param uuid
     * @return
     * @throws UserNotFoundException
     * @throws PasswordNotEqualException
     * @throws UserNotEnabledException
     * @throws PasswordNotChanged
     * @throws NoSuchAlgorithmException
     */
    public boolean confirmPasswordChange(PasswordChangeDTO passwordChangeDTO, String uuid, long time) throws UserNotFoundException, PasswordNotEqualException, UserNotEnabledException, PasswordNotChanged, NoSuchAlgorithmException {

        Optional<UserEntity> optionalUserEntity =
                userRepository.findByUuid(uuid);

        if (optionalUserEntity.isEmpty())
            throw new UserNotFoundException();

        UserEntity userEntity = optionalUserEntity.get();

        String newHash = new HashPwdConfig(passwordChangeDTO.getNewPassword().concat(userEntity.getEmail())).getHash();

        if (time > userEntity.getPasswordChangeRequest()) {
            userEntity.setUuid(null);
            userEntity.setPasswordChangeRequest(null);
            //this.removeUser(userEntity.getEmail());
            userRepository.save(userEntity);
            return false;
        }

        // se password nuova e password confermata non coincidono
        if (!passwordChangeDTO.getNewPassword().equals(passwordChangeDTO.getConfirmNewPassword())) {
            userEntity.setUuid(null);
            userEntity.setPasswordChangeRequest(null);
            //this.removeUser(userEntity.getEmail());
            userRepository.save(userEntity);
            throw new PasswordNotEqualException();
        }


        // se password nuova e vecchia coincidono
        if (userEntity.getPassword().equals(newHash)) {
            String email = userEntity.getEmail();
            String subject = "IBD Tool - Cambio password fallito.";
            String msg = "Gentile utente,\nil cambio password è fallito in quanto la nuova e la vecchia password coincidono.\n" +
                    "Cordiali saluti,\n" +
                    "IBD Tool - Ospedale Mauriziano";

            mailService.sendMail(email, subject, msg);
            userEntity.setUuid(null);
            userEntity.setPasswordChangeRequest(null);
            //this.removeUser(email);
            userRepository.save(userEntity);
            throw new PasswordNotChanged();
        }

        userEntity.setUuid(null);
        userEntity.setPasswordChangeRequest(null);
        userEntity.setPassword(newHash);

        //this.removeUser(userEntity.getEmail());
        userRepository.save(userEntity);

        String subject = "IBD Tool - Cambio password";

        String message = "La tua nuova password per accedere al portale è: " + passwordChangeDTO.getNewPassword() + "\n" +
                "Accedi: " + link + "/login\n" +
                "Cordiali saluti,\n" +
                "IBD Tool - Ospedale Mauriziano";


        mailService.sendMail(userEntity.getEmail(), subject, message);
        return true;
    }

    /**
     * Login service
     * @param loginFormDTO
     * @return
     * @throws UserNotFoundException
     * @throws RegistrationExpiredException
     * @throws UserNotEnabledException
     */
    public EntityDTO login(LoginFormDTO loginFormDTO) throws UserNotFoundException, RegistrationExpiredException, UserNotEnabledException, NoSuchAlgorithmException {

        String hash = new HashPwdConfig(loginFormDTO.getPassword().concat(loginFormDTO.getEmail().toLowerCase())).getHash();
        Optional<UserEntity> optionalUserEntity = userRepository.findByEmailAndPassword(loginFormDTO.getEmail().toLowerCase(), hash);

        if (optionalUserEntity.isEmpty()) {
            throw new UserNotFoundException();
        }

        UserEntity userEntity = optionalUserEntity.get();

        if (!userEntity.isEnabled()) {
            if (userEntity.getRole().getAuthority().equals("DOTTORE")) {
                long time = LocalDateTime.now().atZone(ZoneOffset.UTC).toEpochSecond();
                if (time > userEntity.getRegistrationTime()) {
                    userRepository.delete(userEntity);
                    throw new RegistrationExpiredException();
                } else {
                    throw new UserNotEnabledException();
                }
            }
        }

        // se voluto si può inserire alcune informazioni dell'utente nel jwt
        // modificando il metodo createJwt().

        String jwt = jwtProvider.createJwt(userEntity.getEmail());

        LocalDateTime lastLogin = ZonedDateTime.now(ZoneId.of("Europe/Rome")).toLocalDateTime();
        userEntity.setLastLogin(lastLogin);
        userRepository.save(userEntity);

        if (userEntity.getRole().getAuthority().equals("DOTTORE")) {
            return new EntityDTO(
                    userEntity.getEmail(),
                    userEntity.getName(),
                    userEntity.getSurname(),
                    userEntity.getSSN(),
                    userEntity.getPhoneNumber(),
                    userEntity.getRole().getAuthority(),
                    userEntity.getBirthDate(),
                    userEntity.getBirthPlace(),
                    userEntity.getDoctorID(),
                    null,
                    null,
                    null,
                    jwt,
                    LocalDateTime.now().plusMinutes(15).toEpochSecond(ZoneOffset.UTC),
                    userEntity.getNotifications());
        }

        else if (userEntity.getRole().getAuthority().equals("PAZIENTE")) {
            return new EntityDTO(
                    userEntity.getEmail(),
                    userEntity.getName(),
                    userEntity.getSurname(),
                    userEntity.getSSN(),
                    userEntity.getPhoneNumber(),
                    userEntity.getRole().getAuthority(),
                    userEntity.getBirthDate(),
                    userEntity.getBirthPlace(),
                    null,
                    userEntity.getDoctor(),
                    userEntity.getBodyStat(),
                    userEntity.getQuestionnaires(),
                    jwt,
                    LocalDateTime.now().plusMinutes(15).toEpochSecond(ZoneOffset.UTC),
                    userEntity.getNotifications()
            );
        }

        else {  // is ADMINISTRATOR
            return new EntityDTO(
                    userEntity.getEmail(),
                    userEntity.getName(),
                    userEntity.getSurname(),
                    userEntity.getSSN(),
                    userEntity.getPhoneNumber(),
                    userEntity.getRole().getAuthority(),
                    userEntity.getBirthDate(),
                    userEntity.getBirthPlace(),
                    null,
                    null,
                    null,
                    null,
                    jwt,
                    LocalDateTime.now().plusMinutes(15).toEpochSecond(ZoneOffset.UTC),
                    userEntity.getNotifications()
            );
        }
    }

    /**
     * Get user from the system by SSN
     * @param email
     * @return
     * @throws UserNotFoundException
     */
    public UserEntity getUser(String email) throws UserNotFoundException {
        Optional<UserEntity> optionalUserEntity = userRepository.findByEmail(email);
        if (optionalUserEntity.isEmpty())
            throw new UserNotFoundException();

        return optionalUserEntity.get();
    }

    /**
     * Remove user from the system
     * @param email
     * @throws UserNotFoundException
     */
    public void removeUser(String email) throws UserNotFoundException {
        Optional<UserEntity> toRemove = userRepository.findByEmail(email);

        if (toRemove.isEmpty())
            throw new UserNotFoundException();

        userRepository.delete(toRemove.get());
    }

    /**
     * A doctor can register a new patient, modification applied to doctor
     * @param patientDTO
     * @throws UserNotFoundException
     */
    @Transactional
    public void registerPatient(PatientDTO patientDTO, String ctg) throws IOException, NoSuchAlgorithmException, RequestNotAuthorizedException, QuestionnaireAlreadyCreatedException, QuestionnaireAlreadyCompiledException {

        Optional<UserEntity> optionalDoctor = userRepository.findByDoctorID(patientDTO.getDoctorID());

        if (optionalDoctor.isEmpty())
            throw new UserNotFoundException();


        this.authentication = SecurityContextHolder.getContext().getAuthentication();

        if (!authentication.getPrincipal().toString().equals(optionalDoctor.get().getEmail()))
            throw new RequestNotAuthorizedException();


        Optional<UserEntity> optionalPatient = userRepository.findByEmail(patientDTO.getEmail());

        if (optionalPatient.isPresent())
            throw new UserAlreadyRegisteredException();

        UserEntity doctor = optionalDoctor.get();

        Doctor doctorForPatient = Doctor.builder()
                .email(doctor.getEmail())
                .name(doctor.getName())
                .surname(doctor.getSurname())
                .SSN(doctor.getSSN())
                .phoneNumber(doctor.getPhoneNumber())
                .doctorID(doctor.getDoctorID())
                .build();


        String[] birth = patientDTO.getBirthDate().split("/");
        int birthDay = Integer.parseInt(birth[0]);
        int birthMonth = Integer.parseInt(birth[1]);
        int birthYear = Integer.parseInt(birth[2]);

        LocalDate birthDate = LocalDate.of(birthYear, birthMonth, birthDay);
        String birthPlace = "";
        String phoneNumber = "";
        String password = new HashPwdConfig(patientDTO.getDoctorID().concat(patientDTO.getSSN())).getHash().substring(0, 5);

        BodyStat bodyStat = patientDTO.getBodyStat();
        bodyStat.setNotes(new HashSet<>());
        bodyStat.setTherapy("");

        //System.out.println(patientDTO.getBodyStat());

        //System.out.println(ctg);

        String category = ctg;

        if (category.equals("RANDOM")) {
            category = this.categoryRandomization();
        }

        UserEntity patientEntity = UserEntity.builder()
                .email(patientDTO.getEmail().toLowerCase())
                .name(patientDTO.getName())
                .surname(patientDTO.getSurname())
                .SSN(patientDTO.getSSN())
                .birthDate(birthDate)
                .birthPlace(birthPlace)
                .enabled(false)
                .password(new HashPwdConfig(password.concat(patientDTO.getEmail())).getHash())
                .registrationTime(LocalDateTime.now().atZone(ZoneOffset.UTC).toEpochSecond())
                .phoneNumber(phoneNumber)
                .doctorID(null)
                .role(new Role(UserRole.PAZIENTE))
                .uuid(null)
                .bodyStat(bodyStat)
                .doctor(doctorForPatient)
                .patients(null)
                .questionnaires(new HashSet<>())
                .tokens(new HashSet<>())
                .notifications(new HashSet<>())
                .category(category)
                .build();

        // ADDING QUESTIONNAIRES AND JOBS!!!!!!!!!
        //////////////////////////////////////////


        HashSet<String> questionnaires = new HashSet<>();
        //////////////////////////////////////////

        userRepository.save(patientEntity);

        assignQuestionnairesAndActivateJobs(patientEntity, patientDTO, questionnaires, category, bodyStat);

        userRepository.save(patientEntity); // UPDATE AFTER INSERT

        ArrayList<Patient> patients = doctor.getPatients(); // TODO: using set?

        if (patients == null)
            patients = new ArrayList<>();

        Patient patientForDoctor = Patient.builder()
                .email(patientEntity.getEmail().toLowerCase())
                .name(patientEntity.getName())
                .surname(patientEntity.getSurname())
                .SSN(patientEntity.getSSN())
                .birthDate(patientDTO.getBirthDate())
                .birthPlace(birthPlace)
                .phoneNumber("")
                .bodyStat(patientEntity.getBodyStat())
                .doctor(patientEntity.getDoctor())
                .questionnaires(questionnaires)
                .category(category)
                .build();

        patients.add(patientForDoctor);
        doctor.setPatients(patients);

        //this.removeUser(doctor.getEmail());

        userRepository.save(doctor);

        String subject = "Benvenuto in IBD Tool";
        String message = "Gentile " + patientEntity.getName() + " " + patientEntity.getSurname() +",\n\n" +
                "Benvenuto in IBD Tool!\n" +
                /*"Scarica subito la nostra app dal Play Store! : " + linkApp + "\n" +
                "L'ID del tuo medico è: " + patientDTO.getDoctorID() + "\n" +*/
                "Il link per accedere al sito è: " + link + "\n" +
                "Le tue credenziali per accedere sono:\n" +
                "email: " + patientEntity.getEmail() + "\n" +
                "password generata: " + password + "\n\n" +
                "Ti invitiamo a modificare la tua password al più presto al seguente link:\n" +
                link + "/change-password" + "\n\n" +
                "Cordiali saluti,\n" +
                "IBD Tool - Ospedale Mauriziano";


        mailService.sendMail(patientEntity.getEmail(), subject, message);

        // SEND FIRST NOTIFICATION TO NEW PATIENT REGISTERED

        String subj = "IBD Tool - Aggiorna il tuo profilo";
        String msg = "Benvenuto in IBD Tool.\n" +
                "A breve riceverai un manuale di istruzioni per l'utilizzo.\n" +
                "Nel frattempo aggiorna il tuo profilo dalle impostazioni personali.";

        NotificationRequestDto notificationRequestDto = new NotificationRequestDto();
        notificationRequestDto.setTarget("-");
        notificationRequestDto.setBody(msg);
        LocalDateTime nd = ZonedDateTime.now(ZoneId.of("Europe/Rome")).toLocalDateTime();
        String title = nd.format(DateTimeFormatter.ofPattern("dd/MM/yyyy")) + " " +
                (nd.getHour() < 10 ? ("0" + nd.getHour()) : (nd.getHour())) + ":" +
                (nd.getMinute() < 10 ? ("0" + nd.getMinute()) : (nd.getMinute())) + ":" +
                (nd.getSecond() < 10 ? ("0" + nd.getSecond()) : (nd.getSecond())) +
                " " + subj;
        notificationRequestDto.setTitle(title);
        notificationRequestDto.setTargetEmail(patientEntity.getEmail());

        notificationService.sendPushNotification(notificationRequestDto);

        String sbj = "IBD Tool - Manuale di utilizzo";
        String mail = "Gentile paziente,\nconsulta il nostro manuale di utilizzo al link: " + pathPatient + "\n" +
                "Cordiali saluti,\n" +
                "IBD Tool - Ospedale Mauriziano";


        mailService.sendMail(patientEntity.getEmail(), sbj, mail);
    }

    /**
     * Activated new jobs and assign questionnaires
     * @param patientEntity
     * @param patientDTO
     * @param questionnaires
     * @param category
     * @param bodyStat
     */
    private void assignQuestionnairesAndActivateJobs(UserEntity patientEntity, PatientDTO patientDTO, HashSet<String> questionnaires, String category, BodyStat bodyStat) {
        if (category.equals("TELEMEDICINA")) {
            if (bodyStat.getPathology().equals("Malattia di Crohn")) {
                HashSet<String> everyMonth = new HashSet<>();
                HashSet<String> everyThreeMonth = new HashSet<>();
                HashSet<String> everySixMonth = new HashSet<>();

                // ogni mese
                everyMonth.add("HBI");
                everyMonth.add("MIAH-CD");
                everyMonth.add("IBD-DISK");

                initializeQuestionnaires(patientDTO, questionnaires, everyMonth, everyThreeMonth, everySixMonth);

               // System.out.println(questionnaires);
                patientEntity.setQuestionnaires(questionnaires);

            } else if (bodyStat.getPathology().equals("Colite ulcerosa")) {
                HashSet<String> everyMonth = new HashSet<>();
                HashSet<String> everyThreeMonth = new HashSet<>();
                HashSet<String> everySixMonth = new HashSet<>();

                // ogni mese
                everyMonth.add("PATIENT-SCCAI");
                everyMonth.add("MIAH-UC");
                everyMonth.add("IBD-DISK");

                initializeQuestionnaires(patientDTO, questionnaires, everyMonth, everyThreeMonth, everySixMonth);

                //System.out.println(questionnaires);
                patientEntity.setQuestionnaires(questionnaires);

            } else {

                HashSet<String> everyMonth = new HashSet<>();
                HashSet<String> everyThreeMonth = new HashSet<>();
                HashSet<String> everySixMonth = new HashSet<>();

                initializeQuestionnaires(patientDTO, questionnaires, everyMonth, everyThreeMonth, everySixMonth);

                //System.out.println(questionnaires);
                patientEntity.setQuestionnaires(questionnaires);
            }




        } else if (category.equals("STANDARD")) {
            if (bodyStat.getPathology().equals("Malattia di Crohn")) {
                HashSet<String> everyThreeMonth = new HashSet<>();

                // ogni 3 mese
                everyThreeMonth.add("HBI");
                everyThreeMonth.add("MIAH-CD");
                everyThreeMonth.add("IBD-DISK");

                everyThreeMonth.forEach(q -> {
                    questionnaires.add(q);
                    long timer = 7776000;
                    schedulerService.startTask(patientDTO.getDoctorID(), patientDTO.getSSN(), q, timer);
                });

                //System.out.println(questionnaires);
                patientEntity.setQuestionnaires(questionnaires);

            } else if (bodyStat.getPathology().equals("Colite ulcerosa")) {
                HashSet<String> everyThreeMonth = new HashSet<>();

                // ogni 3 mese
                everyThreeMonth.add("PATIENT-SCCAI");
                everyThreeMonth.add("MIAH-UC");
                everyThreeMonth.add("IBD-DISK");

                everyThreeMonth.forEach(q -> {
                    questionnaires.add(q);
                    long timer = 7776000;
                    schedulerService.startTask(patientDTO.getDoctorID(), patientDTO.getSSN(), q, timer);
                });

                //System.out.println(questionnaires);
                patientEntity.setQuestionnaires(questionnaires);

            } else {
                HashSet<String> everyThreeMonth = new HashSet<>();

                // ogni 3 mesi
                everyThreeMonth.add("IBD-DISK");

                everyThreeMonth.forEach(q -> {
                    questionnaires.add(q);
                    long timer = 7776000;
                    schedulerService.startTask(patientDTO.getDoctorID(), patientDTO.getSSN(), q, timer);
                });

                //System.out.println(questionnaires);
                patientEntity.setQuestionnaires(questionnaires);
            }
        }
    }

    /**
     * method extracted from duplicated code fragments
     * @param patientDTO
     * @param questionnaires
     * @param everyMonth
     * @param everyThreeMonth
     * @param everySixMonth
     */
    private void initializeQuestionnaires(PatientDTO patientDTO, HashSet<String> questionnaires, HashSet<String> everyMonth, HashSet<String> everyThreeMonth, HashSet<String> everySixMonth) {

        everyMonth.forEach(q -> {
            questionnaires.add(q);
            long timer = 2592000;
            schedulerService.startTask(patientDTO.getDoctorID(), patientDTO.getSSN(), q, timer);
        });

        everyThreeMonth.add("IPAQ-SF");
        everyThreeMonth.add("PHQ9");
        everyThreeMonth.add("WPAI-UC");
        everyThreeMonth.add("MMAS8");
        everyThreeMonth.add("TSQM");
        everyThreeMonth.add("PRISM");

        everyThreeMonth.forEach(q -> {
            questionnaires.add(q);
            long timer = 7776000;
            schedulerService.startTask(patientDTO.getDoctorID(), patientDTO.getSSN(), q, timer);
        });

        everySixMonth.add("EQ5D5L");
        everySixMonth.add("PSQI");

        everySixMonth.forEach(q -> {
            questionnaires.add(q);
            long timer = 15552000;
            schedulerService.startTask(patientDTO.getDoctorID(), patientDTO.getSSN(), q, timer);
        });
    }

    /**
     * Patient randomization category
     * @return
     */
    private String categoryRandomization() {

        Optional<ArrayList<UserEntity>> eHealth = userRepository.findAllByCategory("TELEMEDICINA");
        Optional<ArrayList<UserEntity>> traditionalHealth = userRepository.findAllByCategory("STANDARD");

        int eHealthSize = 0;
        int traditionalHealthSize = 0;

        if (eHealth.isPresent()) {
            eHealthSize = eHealth.get().size();
        }

        if (traditionalHealth.isPresent()) {
            traditionalHealthSize = traditionalHealth.get().size();
        }

        long i = System.currentTimeMillis();

        if (eHealthSize < traditionalHealthSize) {
            return "TELEMEDICINA";
        } else if (eHealthSize > traditionalHealthSize) {
            return "STANDARD";
        } else {
            if (i % 2 == 0) {   // RANDOMIZATION ON CURRENT TIMESTAMP
                return "TELEMEDICINA";
            } else {
                return "STANDARD";
            }
        }
    }

    /**
     * Get all patients
     * @param doctorID
     * @return
     * @throws UserNotFoundException
     * @throws PatientsNotFoundException
     */
    public ArrayList<Patient> getAllPatients(String doctorID) throws UserNotFoundException, PatientsNotFoundException, RequestNotAuthorizedException {

        Optional<UserEntity> optionalDoctor = userRepository.findByDoctorID(doctorID);

        if (optionalDoctor.isEmpty())
            throw new UserNotFoundException();

        UserEntity doctor = optionalDoctor.get();

        ArrayList<Patient> patients = doctor.getPatients();

        if (patients == null) {
            throw new PatientsNotFoundException();
        }

        else
            return patients;

    }

    public ArrayList<Patient> getGlobalPatients(String doctorID) throws UserNotFoundException, PatientsNotFoundException {
        Optional<UserEntity> optionalDoctor = userRepository.findByDoctorID(doctorID);

        if (optionalDoctor.isEmpty())
            throw new UserNotFoundException();

        UserEntity doctor = optionalDoctor.get();

        Optional<ArrayList<UserEntity>> optionalPatientEntities = userRepository.findAllByDoctorExists(true);

        if (optionalPatientEntities.isEmpty()) {
            throw new PatientsNotFoundException();
        }

        else {
            ArrayList<Patient> patients = new ArrayList<>();
            for (UserEntity p : optionalPatientEntities.get()) {
                Patient patient = Patient.builder()
                        .birthDate(p.getBirthDate().format(DateTimeFormatter.ofPattern("dd/MM/yyyy")))
                        .birthPlace(p.getBirthPlace())
                        .bodyStat(p.getBodyStat())
                        .category(p.getCategory())
                        .doctor(p.getDoctor())
                        .email(p.getEmail())
                        .name(p.getName())
                        .phoneNumber(p.getPhoneNumber())
                        .questionnaires(p.getQuestionnaires())
                        .SSN(p.getSSN())
                        .surname(p.getSurname())
                        .build();
                patients.add(patient);
            }
            return patients;
        }
    }

    /**
     * Remove patient
     * @param doctorID
     * @param email
     * @throws UserNotFoundException
     */
    public void removePatient(String doctorID, String email) throws UserNotFoundException, RequestNotAuthorizedException {
        Optional<UserEntity> optionalPatient = userRepository.findByEmail(email);
        Optional<UserEntity> optionalDoctor = userRepository.findByDoctorID(doctorID);

        if (optionalPatient.isEmpty())
            throw new UserNotFoundException();

        if (optionalDoctor.isEmpty())
            throw new UserNotFoundException();

        this.authentication = SecurityContextHolder.getContext().getAuthentication();

        if (!authentication.getPrincipal().toString().equals(optionalDoctor.get().getEmail()))
            throw new RequestNotAuthorizedException();

        UserEntity doctor = optionalDoctor.get();
        UserEntity patient = optionalPatient.get();

        ArrayList<Patient> patients = doctor.getPatients(); // TODO: using set?

        Patient patientForDoctor = null;
        for (Patient p : patients) {
            if (p.getEmail().equals(email)) {
                patientForDoctor = p;
            }
        }

        assert patientForDoctor != null;

        Optional<ArrayList<Pending>> optionalPendings = pendingRepository.findByPatientSSN(patientForDoctor.getSSN());
        Optional<ArrayList<Job>> optionalJobs = jobRepository.findAllByIdContaining(patientForDoctor.getSSN());

        if (optionalPendings.isPresent()) {
            ArrayList<Pending> pendingRequestsForPatient = optionalPendings.get();
            pendingRequestsForPatient.forEach(p -> {
                pendingRepository.delete(p);
            });
        }

        if (optionalJobs.isPresent()) {
            ArrayList<Job> jobsForPatient = optionalJobs.get();
            jobsForPatient.forEach(j -> {
                schedulerService.killJobById(j.getId());
            });
        }

        patients.remove(patientForDoctor);
        doctor.setPatients(patients);
        userRepository.delete(patient);
        userRepository.save(doctor);

    }

    /**
     * Stub
     */
    public void logout()  {
    }


    /**
     * Get all jobs for patient
     * @param doctorID
     * @param patientSSN
     * @return
     */
    public ArrayList<Job> getAllJobsForPatient(String doctorID, String patientSSN) throws UserNotFoundException {
        Optional<UserEntity> optionalDoctor = userRepository.findByDoctorID(doctorID);
        Optional<UserEntity> optionalPatient = userRepository.findBySSN(patientSSN);

        if (optionalDoctor.isEmpty())
            throw new UserNotFoundException();

        if (optionalPatient.isEmpty())
            throw new UserNotFoundException();

        UserEntity doctor = optionalDoctor.get();
        UserEntity patient = optionalPatient.get();

        Optional<ArrayList<Job>> optionalJobs = jobRepository.findAllByIdContaining(patientSSN);

        if (optionalJobs.isEmpty())
            return null;
        else {
            return optionalJobs.get();
        }
    }

    /**
     * edit user profile
     * @param editFormDTO
     * @return
     * @throws IOException
     * @throws RequestNotAuthorizedException
     * @throws InterruptedException
     */
    public UserUpdateDTO editProfile(EditFormDTO editFormDTO) throws IOException, RequestNotAuthorizedException, InterruptedException {
        Optional<UserEntity> optionalUserEntity = userRepository.findByEmail(editFormDTO.getEmail());

        if (optionalUserEntity.isEmpty())
            throw new UserNotFoundException();

        this.authentication = SecurityContextHolder.getContext().getAuthentication();

        if (!authentication.getPrincipal().toString().equals(optionalUserEntity.get().getEmail()))
            throw new RequestNotAuthorizedException();

        UserEntity userEntity = optionalUserEntity.get();

        //userRepository.delete(userEntity);

        userEntity.setName(editFormDTO.getName());
        userEntity.setSurname(editFormDTO.getSurname());
        userEntity.setPhoneNumber(editFormDTO.getPhoneNumber());
        userEntity.setBirthPlace(editFormDTO.getBirthPlace());

        String[] birth = editFormDTO.getBirthDate().split("/");
        int birthDay = Integer.parseInt(birth[0]);
        int birthMonth = Integer.parseInt(birth[1]);
        int birthYear = Integer.parseInt(birth[2]);

        LocalDate birthDate = LocalDate.of(birthYear, birthMonth, birthDay);

        userEntity.setBirthDate(birthDate);

        switch (userEntity.getRole().getAuthority()) {
            case "DOTTORE":
                ArrayList<Patient> patients = userEntity.getPatients();

                if (patients != null && patients.size() > 0) {
                    for (Patient patient : patients) {
                        Doctor doctor = Doctor.builder()
                                .doctorID(userEntity.getDoctorID())
                                .email(userEntity.getEmail())
                                .name(userEntity.getName())
                                .surname(userEntity.getSurname())
                                .phoneNumber(userEntity.getPhoneNumber())
                                .SSN(userEntity.getSSN())
                                .build();

                        patient.setDoctor(doctor);

                        Optional<UserEntity> optP = userRepository.findByEmail(patient.getEmail());
                        UserEntity p = optP.orElseThrow();
                        p.setDoctor(doctor);

                        userRepository.save(p);
                    }
                }

                userEntity.setPatients(patients);
                break;

            case "PAZIENTE":

                Doctor d = userEntity.getDoctor();
                Optional<UserEntity> optionalDoctor = userRepository.findByEmail(d.getEmail());
                UserEntity doctor = optionalDoctor.orElseThrow();

                ArrayList<Patient> patientArrayList = doctor.getPatients();

                Patient patientForDoctor = null;
                for (Patient p : patientArrayList) {
                    if (p.getEmail().equals(userEntity.getEmail())) {
                        patientForDoctor = p;
                    }
                }

                patientArrayList.remove(patientForDoctor);

                assert patientForDoctor != null;
                patientForDoctor.setName(editFormDTO.getName());
                patientForDoctor.setSurname(editFormDTO.getSurname());
                patientForDoctor.setPhoneNumber(editFormDTO.getPhoneNumber());
                patientForDoctor.setBirthDate(editFormDTO.getBirthDate());
                patientForDoctor.setBirthPlace(editFormDTO.getBirthPlace());

                patientArrayList.add(patientForDoctor);
                doctor.setPatients(patientArrayList);
                userRepository.save(doctor);

                String body = "Gentile " + doctor.getName() + " " + doctor.getSurname() + "," + " il tuo paziente " + patientForDoctor.getName() + " " + patientForDoctor.getSurname() + " (codice fiscale: " + patientForDoctor.getSSN() + ") ha aggiornato le sue informazioni di contatto.\n";
                String subject = "IBD Tool - Aggiornamento informazioni di " + patientForDoctor.getName() + " " + patientForDoctor.getSurname();

                this.notificationService.sendNotificationToUser(doctor, subject, body);

                break;
        }



        userRepository.save(userEntity);

        return UserUpdateDTO.builder()
                .email(userEntity.getEmail())
                .name(userEntity.getName())
                .surname(userEntity.getSurname())
                .SSN(userEntity.getSSN())
                .role(userEntity.getRole().getAuthority())
                .phoneNumber(userEntity.getPhoneNumber())
                .birthDate(userEntity.getBirthDate())
                .birthPlace(userEntity.getBirthPlace())
                .notifications(userEntity.getNotifications())
                .build();

    }

    /**
     * edit bodyStat of patient
     * @param doctorID
     * @param patientEmail
     * @param bodyStatDTO
     * @return
     * @throws UserNotFoundException
     * @throws RequestNotAuthorizedException
     */
    public BodyStat editPatientClinicalProfile(String doctorID, String patientEmail, BodyStat bodyStatDTO) throws UserNotFoundException, RequestNotAuthorizedException {

        Optional<UserEntity> optionalDoctor = userRepository.findByDoctorID(doctorID);
        Optional<UserEntity> optionalPatient = userRepository.findByEmail(patientEmail);

        if (optionalDoctor.isEmpty())
            throw new UserNotFoundException();

        if (optionalPatient.isEmpty())
            throw new UserNotFoundException();

        this.authentication = SecurityContextHolder.getContext().getAuthentication();

        if (!authentication.getPrincipal().toString().equals(optionalDoctor.get().getEmail()))
            throw new RequestNotAuthorizedException();

        UserEntity doctor = optionalDoctor.get();
        UserEntity patient = optionalPatient.get();

        // updating patient BodyStat and patientForDoctor BodyStat

        if (!patient.getBodyStat().getPathology().equals(bodyStatDTO.getPathology())) {
            Optional<ArrayList<Job>> optionalJobs = jobRepository.findAllByIdContaining(patient.getSSN());

            // Kill old jobs
            if (optionalJobs.isPresent()) {
                ArrayList<Job> jobsForPatient = optionalJobs.get();
                jobsForPatient.forEach(j -> {
                    if (jobKillCondition(j)) {
                        schedulerService.killJobById(j.getId());
                    }
                });
            }

            HashSet<String> questionnaires = new HashSet<>();
            assignQuestionnairesAndActivateJobs(patient, PatientDTO.builder()
                    .SSN(patient.getSSN())
                    .doctorID(doctorID)
                    .build(),
                    questionnaires,
                    patient.getCategory(),
                    bodyStatDTO);
        }

        patient.setBodyStat(bodyStatDTO);
        userRepository.save(patient);

        ArrayList<Patient> patients = doctor.getPatients();

        Patient patientForDoctor = null;
        for (Patient p : patients) {
            if (p.getEmail().equals(patientEmail)) {
                patientForDoctor = p;
            }
        }

        assert patientForDoctor != null;
        patients.remove(patientForDoctor);
        // UPDATE PATIENT INFO
        patientForDoctor.setBodyStat(bodyStatDTO);
        patientForDoctor.setQuestionnaires(patient.getQuestionnaires());

        patients.add(patientForDoctor);

        doctor.setPatients(patients);

        userRepository.save(doctor);

        return bodyStatDTO;
    }

    private boolean jobKillCondition(Job j) {
        String id = j.getId();
        String type = id.split("/")[2];

        switch (type) {
            case "IBD-DISK":
            case "IPAQ-SF":
            case "PHQ9":
            case "WPAI-UC":
            case "MMAS8":
            case "TSQM":
            case "PRISM":
            case "EQ5D5L":
            case "PSQI":
                return false;
            default:
                return true;

        }
    }

    public void removePatientNote(String doctorID, String patientEmail, String note) throws UserNotFoundException, RequestNotAuthorizedException {
        Optional<UserEntity> optionalDoctor = userRepository.findByDoctorID(doctorID);
        Optional<UserEntity> optionalPatient = userRepository.findByEmail(patientEmail);

        if (optionalDoctor.isEmpty())
            throw new UserNotFoundException();

        if (optionalPatient.isEmpty())
            throw new UserNotFoundException();

        this.authentication = SecurityContextHolder.getContext().getAuthentication();

        if (!authentication.getPrincipal().toString().equals(optionalDoctor.get().getEmail()))
            throw new RequestNotAuthorizedException();

        UserEntity doctor = optionalDoctor.get();
        UserEntity patient = optionalPatient.get();

        // updating patient BodyStat and patientForDoctor BodyStat

        BodyStat bodyStat = patient.getBodyStat();
        HashSet<String> notes = bodyStat.getNotes();
        notes.remove(note);
        bodyStat.setNotes(notes);
        patient.setBodyStat(bodyStat);

        userRepository.save(patient);

        ArrayList<Patient> patients = doctor.getPatients();

        Patient patientForDoctor = null;
        for (Patient p : patients) {
            if (p.getEmail().equals(patientEmail)) {
                patientForDoctor = p;
            }
        }

        assert patientForDoctor != null;
        patients.remove(patientForDoctor);
        patientForDoctor.setBodyStat(bodyStat);
        patients.add(patientForDoctor);

        doctor.setPatients(patients);

        userRepository.save(doctor);
    }

    public String keepAlive(String email) throws RequestNotAuthorizedException, UserNotFoundException {

        Optional<UserEntity> optionalUserEntity = userRepository.findByEmail(email);

        if (optionalUserEntity.isEmpty())
            throw new UserNotFoundException();

        UserEntity userEntity = optionalUserEntity.get();

        this.authentication = SecurityContextHolder.getContext().getAuthentication();

        if (!authentication.getPrincipal().toString().equals(userEntity.getEmail()))
            throw new RequestNotAuthorizedException();

        return jwtProvider.createJwt(userEntity.getEmail());
    }

    public UserModel pullUpdates(String email, String ssn) throws UserNotFoundException, RequestNotAuthorizedException {

        Optional<UserEntity> optionalUserEntity = userRepository.findByEmailAndSSN(email, ssn);

        if (optionalUserEntity.isEmpty())
            throw new UserNotFoundException();

        UserEntity userEntity = optionalUserEntity.get();
        this.authentication = SecurityContextHolder.getContext().getAuthentication();

        if (!authentication.getPrincipal().toString().equals(userEntity.getEmail()))
            throw new RequestNotAuthorizedException();

        return new UserModel(
                userEntity.getEmail(),
                userEntity.getName(),
                userEntity.getSurname(),
                userEntity.getSSN(),
                userEntity.getPhoneNumber(),
                userEntity.getRole().getAuthority(),
                userEntity.getBirthDate(),
                userEntity.getBirthPlace(),
                userEntity.getNotifications());
    }

    public Boolean isPatientEnabled(String email) throws UserNotFoundException, RequestNotAuthorizedException {
        Optional<UserEntity> optionalUserEntity = userRepository.findByEmail(email);

        if (optionalUserEntity.isEmpty())
            throw new UserNotFoundException();

        UserEntity userEntity = optionalUserEntity.get();
        this.authentication = SecurityContextHolder.getContext().getAuthentication();

        if (!authentication.getPrincipal().toString().equals(userEntity.getEmail()))
            throw new RequestNotAuthorizedException();

        return userEntity.isEnabled();
    }

    public void enablePatient(String email) throws IOException, RequestNotAuthorizedException, QuestionnaireAlreadyCreatedException, QuestionnaireAlreadyCompiledException {
        Optional<UserEntity> optionalUserEntity = userRepository.findByEmail(email);

        if (optionalUserEntity.isEmpty())
            throw new UserNotFoundException();

        UserEntity userEntity = optionalUserEntity.get();
        this.authentication = SecurityContextHolder.getContext().getAuthentication();

        if (!authentication.getPrincipal().toString().equals(userEntity.getEmail()))
            throw new RequestNotAuthorizedException();

        userEntity.setEnabled(true);
        userRepository.save(userEntity);

        // SENDING FIRST QUESTIONNAIRES AFTER REGISTRATION
        if (userEntity.getBodyStat().getPathology().equals("Malattia di Crohn")) {
            questionnaireService.createAndSend("HBI", userEntity.getDoctor().getDoctorID(), userEntity.getSSN());
            questionnaireService.createAndSend("MIAH-CD", userEntity.getDoctor().getDoctorID(), userEntity.getSSN());
            questionnaireService.createAndSend("IBD-DISK", userEntity.getDoctor().getDoctorID(), userEntity.getSSN());
        } else if (userEntity.getBodyStat().getPathology().equals("Colite ulcerosa")) {
            questionnaireService.createAndSend("PATIENT-SCCAI", userEntity.getDoctor().getDoctorID(), userEntity.getSSN());
            questionnaireService.createAndSend("MIAH-UC", userEntity.getDoctor().getDoctorID(), userEntity.getSSN());
            questionnaireService.createAndSend("IBD-DISK", userEntity.getDoctor().getDoctorID(), userEntity.getSSN());
        } else {
            questionnaireService.createAndSend("IBD-DISK", userEntity.getDoctor().getDoctorID(), userEntity.getSSN());
        }
    }
}
