import { Component, OnInit } from '@angular/core';
import {FormControl, FormGroup, Validators} from '@angular/forms';
import {Observable} from 'rxjs';
import {Patient} from '../../../model/user/Patient';
import {ActivatedRoute, Router} from '@angular/router';
import {Store} from '@ngrx/store';
import {AppState, selectAuthState, selectDoctorState} from '../../../store/app.states';
import {GetPatients} from '../../../store/actions/doctor.actions';
import {ErrorDialogComponent} from '../../user-actions/error-dialog/error-dialog.component';
import {MatSnackBar} from '@angular/material/snack-bar';
import {MatDialog} from '@angular/material/dialog';
import {StatForm} from '../../../model/user/StatForm';
import {DoctorService} from '../../../services/doctor.service';
import * as _moment from 'moment';
// @ts-ignore
import { default as _rollupMoment } from 'moment';
import {MatOptionSelectionChange} from '@angular/material/core';
const moment = _rollupMoment || _moment;

@Component({
  selector: 'app-patient-panel',
  templateUrl: './patient-panel.component.html',
  styleUrls: ['./patient-panel.component.css']
})
export class PatientPanelComponent implements OnInit {

  patientSSN: string;
  getAuthState: Observable<any>;
  getDoctorState: Observable<any>;
  //// user state
  isAuthenticated: boolean;
  user = null;
  errorMessage = null;
  // panel
  patient: Patient;
  title = 'Scheda paziente';
  destination: string;
  editing = false;
  statForm: FormGroup;
  notFocused = false;
  pathologies = ['Malattia di Crohn', 'Colite ulcerosa', 'Altro'];
  otherPathology = false;

  constructor(private router: Router,
              private store: Store<AppState>,
              private route: ActivatedRoute,
              private snackBar: MatSnackBar,
              private dialog: MatDialog,
              private doctorService: DoctorService) {
    this.getAuthState = this.store.select(selectAuthState);
    this.getDoctorState = this.store.select(selectDoctorState);
    this.route.params.subscribe( params => {
      this.patientSSN = params.ssn;
      if (params.back.localeCompare('global') === 0) {
        this.destination = 'patients-global-control-panel';
      } else {
        this.destination = 'patients-control-panel';
      }
    });

    document.addEventListener('notification', () => {
      this.store.dispatch(new GetPatients());
    });

    const patients: Patient[] = JSON.parse(localStorage.getItem('patients'));
    this.patient = patients.find(p => p.ssn.localeCompare(this.patientSSN) === 0);
    this.patient.bodyStat.notes = this.patient.bodyStat.notes.slice().sort((n1, n2) => this.comparing(n1, n2));
    console.log(this.patient);
  }

  ngOnInit(): void {
    this.getAuthState.subscribe((state) => {
      this.isAuthenticated = state.isAuthenticated;
      this.user = JSON.parse(localStorage.getItem('user'));
      this.errorMessage = state.errorMessage;
      console.log(state.user);
    });

    this.statForm = new FormGroup({
      category: new FormControl(this.patient.category, [Validators.required]),
      weight: new FormControl(this.patient.bodyStat.weight, [Validators.required, Validators.min(0)]),
      height: new FormControl(this.patient.bodyStat.height, [Validators.required, Validators.min(0)]),
      pathology: new FormControl(this.patient.bodyStat.pathology, [Validators.required]),
      notes: new FormControl('', []),
      therapy: new FormControl(this.patient.bodyStat.therapy, [])
    });

  }

  selectResult($event: MatOptionSelectionChange, value: string) {
    if ($event.isUserInput) {
      if (value.localeCompare('Altro') !== 0) {
        this.statForm.get('pathology').setValue(value);
        this.otherPathology = false;
      } else {
        this.statForm.get('pathology').setValue('');
        this.otherPathology = true;
      }
      console.log(this.statForm.get('pathology').value);
    }
  }

  comparing(n1: string, n2: string) {
    const first = n1.split(' ');
    const second = n2.split(' ');
    const d1 = moment(first[0] + ' ' + first[1], 'DD/MM/YYYY h:mm:ss').toDate().getTime();
    const d2 = moment(second[0] + ' ' + second[1], 'DD/MM/YYYY h:mm:ss').toDate().getTime();
    return d1 < d2 ? 1 : (d1 > d2 ? -1 : 0);
  }

  calcDate(): string {
    const date = new Date();
    return date.getDate() + '/' +
      (date.getMonth() + 1) + '/' +
      date.getFullYear() + ' ' +
      date.getHours() + ':' +
      date.getMinutes() + ':' +
      date.getSeconds();
  }

  onSubmit(value: any) {
    this.editing = false;
    const notes = this.patient.bodyStat.notes;
    if (value.notes !== null && value.notes !== undefined && value.notes.localeCompare('') !== 0) {
      notes.push(this.calcDate() + ' - ' + value.notes);
    }
    const statForm = new StatForm(value.weight, value.height, value.pathology, notes, value.therapy);
    this.doctorService.editPatientClinicalProfile(this.patient.email, statForm).subscribe(res => {
      if (res.status === 200) {
        this.openSnackBar('Informazioni modificate con successo!', 'Chiudi');
        let patients: Patient[] = JSON.parse(localStorage.getItem('patients'));
        patients = patients.slice().filter(p => p.ssn.localeCompare(this.patientSSN) !== 0);
        // @ts-ignore
        this.patient.bodyStat = res.body;
        patients.push(this.patient);
        localStorage.setItem('patients', JSON.stringify(patients));
        this.statForm.get('notes').reset();
        this.patient.bodyStat.notes = this.patient.bodyStat.notes.slice().sort((n1, n2) => this.comparing(n1, n2));
      }
    },
      error => {
        console.log(error);
        this.openErrorDialog('Qualcosa è andato storto. Le modifiche non verranno salvate.');
      });
  }

  openSnackBar(message: string, action: string) {
    this.snackBar.open(message, action, {
      duration: 2000,
    });
  }

  openErrorDialog(message: string): void {
    const dialogRef = this.dialog.open(ErrorDialogComponent, {
      width: '500px',
      height: '270px',
      data: message
    });

    dialogRef.afterClosed().subscribe(res => {
      console.log('Finestra chiusa');
    });
  }

  modify() {
    this.editing = true;
  }

  deleteNote(n: string) {
    this.doctorService.deleteNote(this.patient.email, n).subscribe(res => {
      if (res.status === 200) {
        this.openSnackBar('Nota eliminata con successo!', 'Chiudi');
        let patients: Patient[] = JSON.parse(localStorage.getItem('patients'));
        patients = patients.slice().filter(p => p.ssn.localeCompare(this.patientSSN) !== 0);
        this.patient.bodyStat.notes = this.patient.bodyStat.notes.slice().filter(note => note.localeCompare(n) !== 0);
        patients.push(this.patient);
        localStorage.setItem('patients', JSON.stringify(patients));
      }
    },
    error => {
      this.openErrorDialog('Si è verificato un errore. Le modifiche non verranno salvate.');
      console.log(error);
    });
  }
}
